#ifndef _ACTIVITY_H
#define _ACTIVITY_H

#define ACTIVITY_OUTPUT 0
#define ACTIVITY_FEATURE 1

/* C API defines */
#define Activity_isActivity_NUM 0
#define Activity_isActivity_RETURN int
#define Activity_isActivity_PROTO (PyObject *obj)

#define Activity_outputFeature_NUM 1
#define Activity_outputFeature_RETURN void
#define Activity_outputFeature_PROTO (PyObject *self, double time_local, double *outputvector, double *featurevector, int DOflag)

#define Activity_nrOutputs_NUM 2
#define Activity_nrOutputs_RETURN int
#define Activity_nrOutputs_PROTO (PyObject *self)

#define Activity_nrFeatures_NUM 3
#define Activity_nrFeatures_RETURN int
#define Activity_nrFeatures_PROTO (PyObject *self)

#define Activity_outputNames_NUM 4
#define Activity_outputNames_RETURN PyObject *
#define Activity_outputNames_PROTO (PyObject *self)

#define Activity_featureNames_NUM 5
#define Activity_featureNames_RETURN PyObject *
#define Activity_featureNames_PROTO (PyObject *self)

#define Activity_featureUnits_NUM 6
#define Activity_featureUnits_RETURN PyObject *
#define Activity_featureUnits_PROTO (PyObject *self)

#define Activity_timeUnit_NUM 7
#define Activity_timeUnit_RETURN PyObject *
#define Activity_timeUnit_PROTO (PyObject *self)

#define Activity_setNonEditable_NUM 8
#define Activity_setNonEditable_RETURN void
#define Activity_setNonEditable_PROTO (PyObject *self)

/* Total number of C API pointers */
#define Activity_API_pointers 9

#ifdef _ACTIVITY_C
/*
==========================================================================================
Structure definitions
==========================================================================================
*/
// Output Objects
typedef double (*OutputFunction)(void *output, double time_local);

typedef struct{
    OutputFunction out_function;
    destructor out_dealloc;
    int type;
    int isFeature;
    double output;
} ActivityOutput;

typedef struct{
    ActivityOutput base;
    double constant;
} ConstantOutput;

typedef struct{
    ActivityOutput base;
    PyObject *tvalues;
    PyObject *fvalues;
} PiecewiseOutput;

typedef struct{
    ActivityOutput base;
    PyObject *tvalues;
    PyObject *fvalues;
    PyObject *slopeendpoints;
    double *b;
    double *c;
    double *d;
    int last;
} CubicSplineOutput;

// Activity Objects
typedef struct{
    PyObject_HEAD
    int nroutputs;
    int nrfeatures;
    int editable; // only editable before added to a simulations object
    PyObject *outputnames;
    PyObject *featurenames;
    PyObject *featureunits;
    ActivityOutput **outputs;
    ActivityOutput **features;
    PyObject *timeunit;
    PyObject *compartment;
} ActivityObject;

/*
==========================================================================================
Function declaration
==========================================================================================
*/
static void
Constant_dealloc(ConstantOutput *self);
static void
Piecewise_dealloc(PiecewiseOutput *self);
static void
CubicInterpolate_dealloc(CubicSplineOutput *self);
static ActivityOutput *
Activity_ConstantOutput(PyObject *constant);
static ActivityOutput *
Activity_PiecewiseConstantOutput(PyObject *tvalues, PyObject *fvalues);
static ActivityOutput *
Activity_PiecewiseLinearOutput(PyObject *tvalues, PyObject *fvalues);
static ActivityOutput *
Activity_CubicSplineOutput(PyObject *tvalues, PyObject *fvalues, PyObject *slopeendpoints);
static ActivityOutput *
Activity_CubicInterpolateSlopeOutput(PyObject *tvalues, PyObject *fvalues, PyObject *slopeendpoints);

static Activity_isActivity_RETURN Activity_isActivity Activity_isActivity_PROTO;
static Activity_outputFeature_RETURN Activity_outputFeature Activity_outputFeature_PROTO;
static Activity_nrOutputs_RETURN Activity_nrOutputs Activity_nrOutputs_PROTO;
static Activity_nrFeatures_RETURN Activity_nrFeatures Activity_nrFeatures_PROTO;
static Activity_outputNames_RETURN Activity_outputNames Activity_outputNames_PROTO;
static Activity_featureNames_RETURN Activity_featureNames Activity_featureNames_PROTO;
static Activity_featureUnits_RETURN Activity_featureUnits Activity_featureUnits_PROTO;
static Activity_timeUnit_RETURN Activity_timeUnit Activity_timeUnit_PROTO;
static Activity_setNonEditable_RETURN Activity_setNonEditable Activity_setNonEditable_PROTO;

#else
/*
==========================================================================================
C_API
==========================================================================================
*/
#define import_Activity() {if (_import_Activity() < 0) {PyErr_Print(); PyErr_SetString(PyExc_ImportError, "MeDigiT._Activity._C_API failed to import"); return NULL; } }
#define Activity_isActivity \
    (*(Activity_isActivity_RETURN (*)Activity_isActivity_PROTO) Activity_API[Activity_isActivity_NUM])
#define Activity_outputFeature \
    (*(Activity_outputFeature_RETURN (*)Activity_outputFeature_PROTO) Activity_API[Activity_outputFeature_NUM])
#define Activity_nrOutputs \
    (*(Activity_nrOutputs_RETURN (*)Activity_nrOutputs_PROTO) Activity_API[Activity_nrOutputs_NUM])
#define Activity_nrFeatures \
    (*(Activity_nrFeatures_RETURN (*)Activity_nrFeatures_PROTO) Activity_API[Activity_nrFeatures_NUM])
#define Activity_outputNames \
    (*(Activity_outputNames_RETURN (*)Activity_outputNames_PROTO) Activity_API[Activity_outputNames_NUM])
#define Activity_featureNames \
    (*(Activity_featureNames_RETURN (*)Activity_featureNames_PROTO) Activity_API[Activity_featureNames_NUM])
 #define Activity_featureUnits \
    (*(Activity_featureUnits_RETURN (*)Activity_featureUnits_PROTO) Activity_API[Activity_featureUnits_NUM])
#define Activity_timeUnit \
    (*(Activity_timeUnit_RETURN (*)Activity_timeUnit_PROTO) Activity_API[Activity_timeUnit_NUM])
#define Activity_setNonEditable \
    (*(Activity_setNonEditable_RETURN (*)Activity_setNonEditable_PROTO) Activity_API[Activity_setNonEditable_NUM])


/* Return -1 on error, 0 on success.
 * PyCapsule_Import will set an exception if there's an error.
 */
static void **Activity_API;
static int
_import_Activity(void)
{
    Activity_API = (void **)PyCapsule_Import("MeDigiT._Activity._C_API", 0);
    return (Activity_API != NULL) ? 0 : -1;
}

#endif
#endif

