#ifndef _MODELS_H
#define _MODELS_H

#define MODEL_STATE 0
#define MODEL_FEATURE 1
#define MODEL_OUTPUT 2
#define MODEL_INPUT 3
#define MODEL_EVENT 4
#define MODEL_PARAMETER 5

/*
==========================================================================================
C_API
==========================================================================================
*/
#define Model_isModel_NUM 0
#define Model_isModel_RETURN int
#define Model_isModel_PROTO (PyObject *obj)

#define Model_modelFunction_NUM 1
#define Model_modelFunction_RETURN ModelFunction *
#define Model_modelFunction_PROTO (PyObject *obj)

#define Model_parameters_NUM 2
#define Model_parameters_RETURN PyObject *
#define Model_parameters_PROTO (PyObject *obj)

#define Model_timeUnit_NUM 3
#define Model_timeUnit_RETURN PyObject *
#define Model_timeUnit_PROTO (PyObject *obj)

#define Model_name_NUM 4
#define Model_name_RETURN PyObject *
#define Model_name_PROTO (PyObject *obj)

#define Model_numberof_NUM 5
#define Model_numberof_RETURN const int *
#define Model_numberof_PROTO (PyObject *obj)

#define Model_inputDependency_NUM 6
#define Model_inputDependency_RETURN const int *
#define Model_inputDependency_PROTO (PyObject *obj)

#define Model_stateValues_NUM 7
#define Model_stateValues_RETURN PyObject *
#define Model_stateValues_PROTO (PyObject *obj)

#define Model_featureNames_NUM 8
#define Model_featureNames_RETURN PyObject *
#define Model_featureNames_PROTO (PyObject *obj)

#define Model_featureUnits_NUM 9
#define Model_featureUnits_RETURN PyObject *
#define Model_featureUnits_PROTO (PyObject *obj)

#define Model_outputNames_NUM 10
#define Model_outputNames_RETURN PyObject *
#define Model_outputNames_PROTO (PyObject *obj)

#define Model_inputNames_NUM 11
#define Model_inputNames_RETURN PyObject *
#define Model_inputNames_PROTO (PyObject *obj)

#define Model_parameterNames_NUM 12
#define Model_parameterNames_RETURN PyObject *
#define Model_parameterNames_PROTO (PyObject *obj)

#define Model_eventNames_NUM 13
#define Model_eventNames_RETURN PyObject *
#define Model_eventNames_PROTO (PyObject *obj)

#define Model_derivativeValues_NUM 14
#define Model_derivativeValues_RETURN PyObject *
#define Model_derivativeValues_PROTO (PyObject *obj)

#define Model_stateNames_NUM 15
#define Model_stateNames_RETURN PyObject *
#define Model_stateNames_PROTO (PyObject *obj)

#define Model_idVector_NUM 16
#define Model_idVector_RETURN PyObject *
#define Model_idVector_PROTO (PyObject *obj)

#define Model_hasAlgebraicEq_NUM 17
#define Model_hasAlgebraicEq_RETURN const int
#define Model_hasAlgebraicEq_PROTO (PyObject *obj)

#define Model_mandatoryInputs_NUM 18
#define Model_mandatoryInputs_RETURN const int *
#define Model_mandatoryInputs_PROTO (PyObject *obj)

/* Total number of C API pointers */
#define Models_C_API_pointers 19

/*
==========================================================================================
M_API
==========================================================================================
*/
#define Model_alloc_NUM 0
#define Model_alloc_RETURN int
#define Model_alloc_PROTO (ModelObject *self)

#define Model_Base_Type_NUM 1
#define Model_Base_Type_RETURN PyTypeObject *

/* Total number of M API pointers */
#define Models_M_API_pointers 2

/*
==========================================================================================
Structure definitions
==========================================================================================
*/
typedef void ModelFunction(double time, double timescale, double *statevalues, double *derivativevalues, double *RESvector, double *parametervector, double *featurevector, double *outputvector, double **inputvector, double *eventvector, int *eventstatus, int DOflag);
typedef void ICFunction(double *icvector, double *dericvector, double *parametervector);

// Model structure
typedef struct {
    ModelFunction * const function;
    ICFunction * const initialcondition;
    const int * const numberof;
    const char ** const statenames;
    const char ** const featurenames;
    const char ** const featureunits;
    const char ** const outputnames;
    const char ** const inputnames;
    const char ** const parameternames;
    const char ** const eventnames;
    const int * const differentialstates;
    const int * const inputdependency;
    const double * const defaultparameters;
    const char * const timeunit;
    const int has_algebraic_eq;
    const int * const mandatoryinputs; // 0 = mandatory, 1 = non-mandatory, 2 = non-mandatory - zero default value
}ModelStructure;

// Model object
typedef struct {
    PyObject_HEAD
    ModelStructure *model; // reference to model structure
    PyObject *name;
    PyObject *compartment;
    PyObject *statevalues;
    PyObject *derivativevalues;
    PyObject *idvector;
    PyObject *parametervalues;
    PyObject *statenames;
    PyObject *featurenames;
    PyObject *featureunits;
    PyObject *outputnames;
    PyObject *inputnames;
    PyObject *parameternames;
    PyObject *eventnames;
    PyObject *timeunit;
}ModelObject;

#if defined(_MODELS_C)
/*
==========================================================================================
Function declaration
==========================================================================================
*/

#elif defined(MeDigiT_MODEL)
/*
==========================================================================================
M_API
==========================================================================================
*/
#define import_Models() {if (_import_Models() < 0) {PyErr_Print(); PyErr_SetString(PyExc_ImportError, "MeDigiT._Models._M_API failed to import"); return NULL; } }
#define Model_alloc \
    (*(Model_alloc_RETURN (*)Model_alloc_PROTO) Models_API[Model_alloc_NUM])
#define Model_Base_Type \
    ((Model_Base_Type_RETURN) Models_API[Model_Base_Type_NUM])

static void **Models_API;
static int
_import_Models(void)
{
    Models_API = (void **)PyCapsule_Import("MeDigiT._Models._M_API", 0);
    return (Models_API != NULL) ? 0 : -1;
}

#else
/*
==========================================================================================
C_API
==========================================================================================
*/

#define import_Models() {if (_import_Models() < 0) {PyErr_Print(); PyErr_SetString(PyExc_ImportError, "MeDigiT._Models._C_API failed to import"); return NULL; } }
#define Model_isModel \
    (*(Model_isModel_RETURN (*)Model_isModel_PROTO) Models_API[Model_isModel_NUM])
#define Model_modelFunction \
    (*(Model_modelFunction_RETURN (*)Model_modelFunction_PROTO) Models_API[Model_modelFunction_NUM])
#define Model_parameters \
    (*(Model_parameters_RETURN (*)Model_parameters_PROTO) Models_API[Model_parameters_NUM])
#define Model_timeUnit \
    (*(Model_timeUnit_RETURN (*)Model_timeUnit_PROTO) Models_API[Model_timeUnit_NUM])
#define Model_name \
    (*(Model_name_RETURN (*)Model_name_PROTO) Models_API[Model_name_NUM])
#define Model_numberof \
    (*(Model_numberof_RETURN (*)Model_numberof_PROTO) Models_API[Model_numberof_NUM])
#define Model_inputDependency \
    (*(Model_inputDependency_RETURN (*)Model_inputDependency_PROTO) Models_API[Model_inputDependency_NUM])
#define Model_stateValues \
    (*(Model_stateValues_RETURN (*)Model_stateValues_PROTO) Models_API[Model_stateValues_NUM])
#define Model_featureNames \
    (*(Model_featureNames_RETURN (*)Model_featureNames_PROTO) Models_API[Model_featureNames_NUM])
#define Model_featureUnits \
    (*(Model_featureUnits_RETURN (*)Model_featureUnits_PROTO) Models_API[Model_featureUnits_NUM])
#define Model_outputNames \
    (*(Model_outputNames_RETURN (*)Model_outputNames_PROTO) Models_API[Model_outputNames_NUM])
#define Model_inputNames \
    (*(Model_inputNames_RETURN (*)Model_inputNames_PROTO) Models_API[Model_inputNames_NUM])
#define Model_parameterNames \
    (*(Model_parameterNames_RETURN (*)Model_parameterNames_PROTO) Models_API[Model_parameterNames_NUM])
#define Model_eventNames \
    (*(Model_eventNames_RETURN (*)Model_eventNames_PROTO) Models_API[Model_eventNames_NUM])
#define Model_derivativeValues \
    (*(Model_derivativeValues_RETURN (*)Model_derivativeValues_PROTO) Models_API[Model_derivativeValues_NUM])
#define Model_stateNames \
    (*(Model_stateNames_RETURN (*)Model_stateNames_PROTO) Models_API[Model_stateNames_NUM])
#define Model_idVector \
    (*(Model_idVector_RETURN (*)Model_idVector_PROTO) Models_API[Model_idVector_NUM])
#define Model_hasAlgebraicEq \
    (*(Model_hasAlgebraicEq_RETURN (*)Model_hasAlgebraicEq_PROTO) Models_API[Model_hasAlgebraicEq_NUM])
#define Model_mandatoryInputs \
    (*(Model_mandatoryInputs_RETURN (*)Model_mandatoryInputs_PROTO) Models_API[Model_mandatoryInputs_NUM])

/* Return -1 on error, 0 on success.
 * PyCapsule_Import will set an exception if there's an error.
 */
static void **Models_API;
static int
_import_Models(void)
{
    Models_API = (void **)PyCapsule_Import("MeDigiT._Models._C_API", 0);
    return (Models_API != NULL) ? 0 : -1;
}

#endif
#endif

