#define _SIMULATION_C
#include <Python.h>
#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
#include <numpy/arrayobject.h>

#include "_StringList.h"
#include "_Models.h"
#include "_Activity.h"
#include "pyarraymacros.h"
#include "timescales.h"
#include "SUNDIALSflags55.h"
#include "SUNDIALSinterface55.h"

#define _SIMULATION_C
#include "_Simulation.h"

#define SIMULATION_STATE 0
#define SIMULATION_FEATURE 1
#define SIMULATION_OUTPUT 2
#define SIMULATION_INPUT 3
#define SIMULATION_EVENT 4
#define SIMULATION_PARAMETER 5
#define SIMULATION_MODEL 6
#define SIMULATION_ACTIVITY 7

static double default_0_input = 0.0;
/*
========================================================================================================================
Helper functions
========================================================================================================================
*/
// Simulations method
void model(void *simData, double time_local, double *statevalues, double *derivativevalues, double *RESvector, double *featurevector, int DOflag, int timeindex, double *eventvector, int *eventstatus){
    SimulationObject *self;
    SimulationModel *mod;
    SimulationActivity *act;
    self = (SimulationObject *) simData;
    int k, *offset;
    double *parametervalues;
    parametervalues = PYDATA(self->parametervalues);

    // Activities
    for(k = 0; k < self->numberof[SIMULATION_ACTIVITY]; k++){
        act = &self->activities[k];
        offset = act->offset;
        Activity_outputFeature(act->activityObject, A_SCALE * time_local, A_OUTPUTVECTOR, A_FEATUREVECTOR, DOflag);
    }

    // Models
    // Update outputs
    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){
        mod = &self->models[k];
        offset = mod->offset;
        if(HASOUTPUT)
            MODELFUNCTION(SCALE * time_local, SCALE, STATEVECTOR, DERIVATEVECTOR, NULL, PARAMETERVECTOR, NULL, OUTPUTVECTOR, INPUTVECTOR, NULL, NULL, DOFLAG_OUTPUT);
    }
    // do what is ask for
    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){
        mod = &self->models[k];
        offset = mod->offset;
        MODELFUNCTION(SCALE * time_local, SCALE, STATEVECTOR, DERIVATEVECTOR, RESIDUALVECTOR, PARAMETERVECTOR, FEATUREVECTOR, OUTPUTVECTOR, INPUTVECTOR, EVENTVECTOR, EVENTSTATUS, DOflag);
    }
}

static int
Simulation_outputOwner(SimulationObject *self, int output){
    int k;
    if(output < 0) // default input
        return -1;
    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){
        if(output < self->models[k].offset[MODEL_OUTPUT] + Model_numberof(self->models[k].modelObject)[MODEL_OUTPUT]
                && output >= self->models[k].offset[MODEL_OUTPUT])
            return k;
    }
    return -2; // activity output
}

static int
Simulation_inputOwner(SimulationObject *self, int input){
    int k;
    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){
        if(input < self->models[k].offset[MODEL_INPUT] + Model_numberof(self->models[k].modelObject)[MODEL_INPUT]
                && input >= self->models[k].offset[MODEL_INPUT])
            return k;
    }
    return -1; // unknown input
}

static int
Simulation_determineExOrder(SimulationObject *self, int modIndex, int *modStat){
    SimulationModel *model;
    const int *inDep;
    int tmp, k, maxExOrd, output;

    if(modIndex < 0)// default_input = -1 or activity = -2
        return 0;
    
    model = &self->models[modIndex];
    if(modStat[modIndex] == -1) // model execution order is being calculated
        return -1; //fail
    else if(modStat[modIndex] == 1) // model execution order has already being calculated
        return model->exOrder;
    else
        modStat[modIndex] = -1;

    // no input dependency
    inDep = Model_inputDependency(model->modelObject);
    if(inDep[0] == 0){
        model->exOrder = 0;
        modStat[modIndex] = 1;
        return model->exOrder;
    }

    // go through input dependencies
    maxExOrd = -1;
    for(k = 1; k <= inDep[0]; k++){
        output = self->inputmap[model->offset[MODEL_INPUT] + inDep[k]];
        tmp = Simulation_determineExOrder(self, Simulation_outputOwner(self, output), modStat);
        if(tmp < 0){ // fail
            return -1;
        }
        if(tmp > maxExOrd)
            maxExOrd = tmp;
    }

    model->exOrder = maxExOrd + 1;
    modStat[modIndex] = 1;
    return model->exOrder;
}

static int
Simulation_compare(const void *p, const void *q){
    const SimulationModel *x, *y;
    x = (const SimulationModel *) p;
    y = (const SimulationModel *) q;

    if(x->exOrder < y->exOrder)
        return -1;
    else if(x->exOrder > y->exOrder)
        return 1;
    return 0;
}

static int
Simulation_CheckSharedVariables(SimulationObject *self){
    PyObject *indexFunc, *ret, *tmp;
    SimulationModel *inputowner;
    int k, *modStat, mandatory;

    self->sharedvariablescheck = 0;

    indexFunc = PyUnicode_FromString("index");
    for(k = 0; k < self->numberof[SIMULATION_INPUT]; k++){
        tmp = PyList_GetItem(self->inputnames, k);
        ret = PyObject_CallMethodObjArgs((PyObject *) self->outputnames, indexFunc, tmp, NULL);
        if(PyErr_Occurred()){// not in list
            inputowner = &self->models[Simulation_inputOwner(self, k)];
            // check if input is mandatory
            mandatory = Model_mandatoryInputs(inputowner->modelObject)[k - inputowner->offset[MODEL_INPUT]]; // 0 mandatory, 1 non-mandatory, 2 non-mandatory - default value 0
            if(mandatory == 0){
                PyErr_SetObject(PyExc_ValueError, 
                    PyUnicode_FromFormat("Model '%S' is missing mandatory input '%S' which is not among the shared variables", 
                                            Model_name(inputowner->modelObject), tmp));
                Py_DECREF(indexFunc);
                return 0;
            }
            else if(mandatory == 1){ // non-mandatory
                self->inputmap[k] = -1;
                self->inputptr[k] = NULL; // default input value used
            }
            else{ // non-mandatory - default value 0
                self->inputmap[k] = -1;
                self->inputptr[k] = &default_0_input; // default 0 input value used
            }
            PyErr_Clear();
        }
        else{
            self->inputmap[k] = PyLong_AsLong(ret); // map output index to input
            self->inputptr[k] = &self->outputbuffer[self->inputmap[k]];
        }
    }
    Py_DECREF(indexFunc);

    modStat = PyMem_Calloc(self->numberof[SIMULATION_MODEL], sizeof(int));  // model status: -1 execution order is being calculated,
                                                            //  0 no action taken, 1 execution order calculated
    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){            
        if(modStat[k] == 0){
            if(Simulation_determineExOrder(self, k, modStat) < 0){
                PyMem_Free(modStat);
                PyErr_SetObject(PyExc_ValueError, 
                    PyUnicode_FromFormat("Cannot resolve execution dependency for model %S", Model_name(self->models[k].modelObject)));
                return 0;
            }
        }
        else if(modStat[k] == -1){
            printf("%s\n", "Something went wrong...");
        }
    }
    self->sharedvariablescheck = 1;
    PyMem_Free(modStat);

    // sort models
    qsort((void *) self->models, self->numberof[SIMULATION_MODEL], sizeof(SimulationModel), Simulation_compare);

    return 1;
}

static void 
Simulation_updateSimulationData(SimulationObject *self, SUNDIALS_SimData *simdata){
    npy_intp dims[2];//, *currentDims;
    PyObject *tmp;

    dims[0] = simdata->nreventstriggered;
    dims[1] = self->numberof[SIMULATION_EVENT];

    // eventtimedata
    tmp = self->eventtimedata;
    self->eventtimedata = PyArray_SimpleNewFromData(1,dims,NPY_DOUBLE,(void *) simdata->eventtimedata);
    PyArray_ENABLEFLAGS((PyArrayObject *) self->eventtimedata, NPY_ARRAY_OWNDATA);
    Py_DECREF(tmp);

    // eventstatusdata
    tmp = self->eventstatusdata;
    self->eventstatusdata = PyArray_SimpleNewFromData(2,dims,NPY_BOOL,(void *) simdata->eventstatusdata);
    PyArray_ENABLEFLAGS((PyArrayObject *) self->eventstatusdata, NPY_ARRAY_OWNDATA);
    Py_DECREF(tmp);

    // featuredata
    dims[0] = PYSIZE(self->timevector);
    dims[1] = self->numberof[SIMULATION_FEATURE];
    tmp = self->featuredata;
    self->featuredata = PyArray_SimpleNewFromData(2,dims,NPY_DOUBLE,(void *) simdata->featuredata);
    PyArray_ENABLEFLAGS((PyArrayObject *) self->featuredata, NPY_ARRAY_OWNDATA);
    Py_DECREF(tmp);
}

static int 
Simulation_InitSimulation(SimulationObject *self){
    // check shared variables
    if(!self->sharedvariablescheck){
        if(!Simulation_CheckSharedVariables(self)){
            return -1;
        }
    }
    // check timevector
    if(PYSIZE(self->timevector) <= 1){
        PyErr_SetString(PyExc_AttributeError, "Not enough timepoints given");
        return -1;
    }

    return 0;
}

static double 
Simulation_lookupTimeScale(const char *timeunit, TimeScale *table){
    while(table->name){
        if(!strcmp(table->name, timeunit))
            return table->scale;
        table++;
    }
    return 0;
}

static const char* 
Simulation_lookupTimeUnit(double scale, TimeScale *table){
    while(table->name){
        if(table->scale == scale)
            return table->name;
        table++;
    }
    return NULL;
}

static int
Simulation_updateModelActivityScale(SimulationObject *self){
    int k;
    double scale;

    // models
    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){
        scale = Simulation_lookupTimeScale(PyUnicode_AsUTF8(Model_timeUnit(self->models[k].modelObject)), timeScaleData);
        if(!scale){
            PyErr_SetObject(PyExc_ValueError, PyUnicode_FromFormat("Incorrect time unit '%S' given in model '%S'.", 
                                                                        Model_timeUnit(self->models[k].modelObject), 
                                                                            Model_name(self->models[k].modelObject)));
            return -1;
        }
        self->models[k].scale = self->scale / scale;
    }
    // activities
    for(k = 0; k < self->numberof[SIMULATION_ACTIVITY]; k++){
        scale = Simulation_lookupTimeScale(PyUnicode_AsUTF8(Activity_timeUnit(self->activities[k].activityObject)), timeScaleData);
        if(!scale){
            PyErr_SetObject(PyExc_ValueError, PyUnicode_FromFormat("Incorrect time unit '%S' in activity nr. %d.", 
                                                                        Activity_timeUnit(self->activities[k].activityObject), 
                                                                            k + 1));
            return -1;
        }
        self->activities[k].scale = self->scale / scale;
    }
    return 0;
}

static int
Simulation_updateDerivativeTimeScale(SimulationObject *self, double old_scale){
    int k;
    double conv_factor, *derivativevalues;

    conv_factor = self->scale / old_scale;
    derivativevalues = PYDATA(self->derivativevalues);
    for(k = 0; k < self->numberof[SIMULATION_STATE]; k++){
        derivativevalues[k] *= conv_factor;
    }

    return 0;
}

static int 
Simulation_CheckActivities(PyObject *activities){
    int k;
    if(Activity_isActivity(activities))
        return 1;
    if(!PyList_Check(activities))
        return -1;
    for(k = 0; k < Py_SIZE(activities); k++){
        if(!Activity_isActivity(PyList_GetItem(activities,k)))
            return -1;
    }
    return Py_SIZE(activities);
}

static int 
Simulation_CheckModels(PyObject *models){
    int k;
    if(Model_isModel(models))
        return 1;
    if(!PyList_Check(models))
        return -1;
    for(k = 0; k < Py_SIZE(models); k++){
        if(!Model_isModel(PyList_GetItem(models,k)))
            return -1;
    }
    return Py_SIZE(models);
}

static int 
Simulation_SetAttributeNames(SimulationObject *self){
    SimulationModel *mod;
    SimulationActivity *act;
    PyObject *obj;
    int k,tmp; 
    const int *numberof;
    // Models
    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){
        mod = &self->models[k];
        obj = mod->modelObject;
        numberof = Model_numberof(obj);
        // featurenames
        if(PyList_SetSlice(self->featurenames,mod->offset[MODEL_FEATURE],mod->offset[MODEL_FEATURE] + numberof[MODEL_FEATURE],Model_featureNames(obj)) < 0)
            return -1;
        // featureunits
        if(PyList_SetSlice(self->featureunits,mod->offset[MODEL_FEATURE],mod->offset[MODEL_FEATURE] + numberof[MODEL_FEATURE],Model_featureUnits(obj)) < 0)
            return -1;
        // outputnames
        if(PyList_SetSlice(self->outputnames,mod->offset[MODEL_OUTPUT],mod->offset[MODEL_OUTPUT] + numberof[MODEL_OUTPUT],Model_outputNames(obj)) < 0)
            return -1;
        // inputnames
        if(PyList_SetSlice(self->inputnames,mod->offset[MODEL_INPUT],mod->offset[MODEL_INPUT] + numberof[MODEL_INPUT],Model_inputNames(obj)) < 0)
            return -1;
        // parameternames
        if(PyList_SetSlice(self->parameternames,mod->offset[MODEL_PARAMETER],mod->offset[MODEL_PARAMETER] + numberof[MODEL_PARAMETER],Model_parameterNames(obj)) < 0)
            return -1;
        // statenames
        if(PyList_SetSlice(self->statenames,mod->offset[MODEL_STATE],mod->offset[MODEL_STATE] + numberof[MODEL_STATE],Model_stateNames(obj)) < 0)
            return -1;
        // eventnames
        if(PyList_SetSlice(self->eventnames,mod->offset[MODEL_EVENT],mod->offset[MODEL_EVENT] + numberof[MODEL_EVENT],Model_eventNames(obj)) < 0)
            return -1;
    }
    // Activities
    for(k = 0; k < self->numberof[SIMULATION_ACTIVITY]; k++){
        act = &self->activities[k];
        obj = act->activityObject;
        tmp = Activity_nrFeatures(obj);
        if(PyList_SetSlice(self->featurenames,act->offset[ACTIVITY_FEATURE],act->offset[ACTIVITY_FEATURE] + tmp,Activity_featureNames(obj)) < 0)
            return -1;
        if(PyList_SetSlice(self->featureunits,act->offset[ACTIVITY_FEATURE],act->offset[ACTIVITY_FEATURE] + tmp,Activity_featureUnits(obj)) < 0)
            return -1;
        if(PyList_SetSlice(self->outputnames,act->offset[ACTIVITY_OUTPUT],act->offset[ACTIVITY_OUTPUT] + Activity_nrOutputs(obj),Activity_outputNames(obj)) < 0)
            return -1;

    }
    return 0;
}

static void 
Simulation_idVectorAlgebraicEqs(SimulationObject *self){
    SimulationModel *mod;
    int k;
    const int *numberof;

    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){
        mod = &self->models[k];
        numberof = Model_numberof(mod->modelObject);
        memcpy(PYDATA(self->idvector) + mod->offset[MODEL_STATE], 
                PYDATA(Model_idVector(mod->modelObject)),
                    numberof[MODEL_STATE] * sizeof(double));
        // has algebraic states
        if(Model_hasAlgebraicEq(mod->modelObject))
            self->has_algebraic_eq = 1;
    }
}

/*
========================================================================================================================
SETTERS AND GETTERS
========================================================================================================================
*/
static PyObject *
Simulation_GetTimeVector(SimulationObject *self, void *closure)
{
    Py_INCREF(self->timevector);
    return self->timevector;
}

static int
Simulation_SetTimeVector(SimulationObject *self, PyObject *value, void *closure)
{
    PyObject *tmp;
    if (value == NULL) {
        PyErr_SetString(PyExc_TypeError, "Cannot delete the timevector attribute");
        return -1;
    }

    value = PyArray_FROM_OTF(value, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if (!value || PyArray_NDIM(PYARRAY(value)) != 1) {
        PyErr_SetString(PyExc_TypeError,
                        "The given value does not match the timevector attribute type");
        Py_XDECREF(value);
        return -1;
    }
    tmp = self->timevector;
    self->timevector = value;
    Py_DECREF(tmp);
    return 0;
}

static PyObject *
Simulation_GetFeatureData(SimulationObject *self, void *closure)
{
    Py_INCREF(self->featuredata);
    return self->featuredata;
}

static PyObject *
Simulation_GetEventTimes(SimulationObject *self, void *closure)
{
    Py_INCREF(self->eventtimedata);
    return self->eventtimedata;
}

static PyObject *
Simulation_GetEventStatus(SimulationObject *self, void *closure)
{
    Py_INCREF(self->eventstatusdata);
    return self->eventstatusdata;
}

// Attributes
static PyObject *
Simulation_GetFeatureNames(SimulationObject *self, void *closure)
{
    Py_INCREF(self->featurenames);
    return self->featurenames;
}

static PyObject *
Simulation_GetFeatureUnits(SimulationObject *self, void *closure)
{   
    
    Py_INCREF(self->featureunits);
    return self->featureunits;
}

static PyObject *
Simulation_GetOutputNames(SimulationObject *self, void *closure)
{
    Py_INCREF(self->outputnames);
    return self->outputnames;
}

static PyObject *
Simulation_GetInputNames(SimulationObject *self, void *closure)
{
    Py_INCREF(self->inputnames);
    return self->inputnames;
}

static PyObject *
Simulation_GetParameterNames(SimulationObject *self, void *closure)
{
    Py_INCREF(self->parameternames);
    return self->parameternames;
}

static PyObject *
Simulation_GetStateNames(SimulationObject *self, void *closure)
{
    Py_INCREF(self->statenames);
    return self->statenames;
}

static PyObject *
Simulation_GetParameterValues(SimulationObject *self, void *closure)
{
    Py_INCREF(self->parametervalues);
    return self->parametervalues;
}

static int
Simulation_SetParameterValues(SimulationObject *self, PyObject *value, void *closure)
{
    PyObject *tmp;

    value = PyArray_FROM_OTF(value, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if(!value || PYSIZE(value) != self->numberof[SIMULATION_PARAMETER]){
        PyErr_SetString(PyExc_TypeError,
                        "The given value does not match the parametervalues attribute type");
        Py_XDECREF(value);
        return -1;
    }
    tmp = self->parametervalues;
    self->parametervalues = value;
    Py_DECREF(tmp);

    return 0;
}

static PyObject *
Simulation_GetStateValues(SimulationObject *self, void *closure)
{
    Py_INCREF(self->statevalues);
    return self->statevalues;
}

static int
Simulation_SetStateValues(SimulationObject *self, PyObject *value, void *closure)
{
    PyObject *tmp;

    value = PyArray_FROM_OTF(value, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if(!value || PYSIZE(value) != self->numberof[SIMULATION_STATE]){
        PyErr_SetString(PyExc_TypeError,
                        "The given value does not match the statevalues attribute type");
        Py_XDECREF(value);
        return -1;
    }
    tmp = self->statevalues;
    self->statevalues = value;
    Py_DECREF(tmp);

    return 0;
}

static PyObject *
Simulation_GetDerivativeValues(SimulationObject *self, void *closure)
{
    Py_INCREF(self->derivativevalues);
    return self->derivativevalues;
}

static int
Simulation_SetDerivativeValues(SimulationObject *self, PyObject *value, void *closure)
{
    PyObject *tmp;
    //double *data;
    //int k;

    value = PyArray_FROM_OTF(value, NPY_DOUBLE, NPY_ARRAY_DEFAULT | NPY_ARRAY_ENSURECOPY);
    if(!value || PYSIZE(value) != self->numberof[SIMULATION_STATE]){
        PyErr_SetString(PyExc_TypeError,
                        "The given value does not match the derivativevalues attribute type");
        Py_XDECREF(value);
        return -1;
    }
    tmp = self->derivativevalues;
    self->derivativevalues = value;
    Py_DECREF(tmp);

    return 0;
}

static PyObject *
Simulation_GetEventNames(SimulationObject *self, void *closure)
{
    Py_INCREF(self->eventnames);
    return self->eventnames;
}

static int
Simulation_getTimeScale(SimulationObject *self, PyObject *value)
{
    double scale;
    const char *timeunit = NULL;

    if(!PyUnicode_Check(value)){
        PyErr_SetString(PyExc_TypeError,
                        "Time unit has to be a unicode object");
        return -1;
    }
    timeunit = PyUnicode_AsUTF8(value);
    if(!timeunit)
        return -1;

    scale = Simulation_lookupTimeScale(timeunit, timeScaleData);
    if(scale > 0){
        return scale;
    } else{
        PyErr_SetObject(PyExc_TypeError, 
                    PyUnicode_FromFormat("Incorrect time unit given: '%S'", value));
        return -1;
    }
}

static int 
Simulation_SetTimeUnit(SimulationObject *self, PyObject *value, void *closure)
{
    double scale, old_scale;

    scale = Simulation_getTimeScale(self, value);
    if(scale < 0) // error occured
        return -1;
    else{
        if(scale != self->scale){ // is it a new scale
            old_scale = self->scale;
            self->scale = scale;
            if(Simulation_updateModelActivityScale(self) < 0 || Simulation_updateDerivativeTimeScale(self, old_scale) < 0)
                return -1;
        }
        return 0;
    }
}

static PyObject *
Simulation_GetTimeUnit(SimulationObject *self, void *closure)
{
    const char *timeunit;

    timeunit = Simulation_lookupTimeUnit(self->scale, timeScaleData);
    if(timeunit){
        return PyUnicode_FromString(timeunit);
    } else{
        return PyUnicode_FromString("Undefined-time-scale");
    }
}

static PyObject *
Simulation_GetOptions(SimulationObject *self, void *closure)
{
    return PyDict_Copy(self->options);
}

static int
Simulation_SetOptions(SimulationObject *self, PyObject *value, void *closure)
{
    PyObject *tmp;

    if(!PyDict_Check(value)){
        PyErr_SetString(PyExc_TypeError,
                        "Options needs to be a dictionary");
        return -1;
    }
    tmp = self->options;
    Py_INCREF(value);
    self->options = value;
    Py_DECREF(tmp);

    // update sundials options
    if(self->sundials->setoptions(self->sundials, self->options) < 0)
        return -1;

    return 0;
}


static PyGetSetDef Simulation_getsetters[] = {
    // Simulation attributes
    {"featurenames", (getter) Simulation_GetFeatureNames, NULL,
    "Simulation feature names", NULL},
    {"featureunits", (getter) Simulation_GetFeatureUnits, NULL,
    "Simulation feature units", NULL},
    {"sharedoutputs", (getter) Simulation_GetOutputNames, NULL,
    "Simulation output names", NULL},
    {"sharedinputs", (getter) Simulation_GetInputNames, NULL,
    "Simulation input names", NULL},
    {"parameternames", (getter) Simulation_GetParameterNames, NULL,
    "Simulation parameter names", NULL},
    {"statenames", (getter) Simulation_GetStateNames, NULL,
    "Simulation state names", NULL},
    {"eventnames", (getter) Simulation_GetEventNames, NULL,
    "Simulation event names", NULL},
    // simulation data
    {"timevector", (getter) Simulation_GetTimeVector, (setter) Simulation_SetTimeVector,
     "Simulation time vector", NULL},
    {"featuredata", (getter) Simulation_GetFeatureData, NULL,
     "Simulation feature data", NULL},
    {"eventtimes", (getter) Simulation_GetEventTimes, NULL,
    "Simulation event times", NULL},
    {"eventstatus", (getter) Simulation_GetEventStatus, NULL,
    "Simulation events' status flag", NULL},
    {"statevalues", (getter) Simulation_GetStateValues, (setter) Simulation_SetStateValues,
    "Simulation state values", NULL},
    {"derivativevalues", (getter) Simulation_GetDerivativeValues, (setter) Simulation_SetDerivativeValues,
    "Simulation derivative values", NULL},
    {"parametervalues", (getter) Simulation_GetParameterValues, (setter) Simulation_SetParameterValues,
    "Simulation parameter values", NULL},
    {"timeunit", (getter) Simulation_GetTimeUnit, (setter) Simulation_SetTimeUnit,
    "Simulation time unit", NULL},
    {"options", (getter) Simulation_GetOptions, (setter) Simulation_SetOptions,
    "Simulation options", NULL},
    {NULL}  /* Sentinel */
};

/*
========================================================================================================================
PyMethods
========================================================================================================================
*/
static PyObject *
Simulation_ResetStatesDerivatives(SimulationObject *self, PyObject *args)
{
    SimulationModel *mod;
    int k,l,offset,nrstates;
    double *from, *to;

    to = PYDATA(self->derivativevalues);

    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){
        mod = &self->models[k];
        offset = mod->offset[MODEL_STATE];
        nrstates = Model_numberof(mod->modelObject)[MODEL_STATE];
        // statevalues
        memcpy(PYDATA(self->statevalues) + offset, 
                PYDATA(Model_stateValues(mod->modelObject)),
                    nrstates * sizeof(double));
        // derivativevalues
        from = PYDATA(Model_derivativeValues(mod->modelObject));
        for(l = 0; l < nrstates; l++)
            to[l + offset] = SCALE * from[l];
    }

    Py_RETURN_NONE;
}

static PyObject *
Simulation_UpdateModelStatevalues(SimulationObject *self, PyObject *args)
{
    SimulationModel *mod;
    int k,l,offset,nrstates;
    double scale_inv,*from, *to;

    from = PYDATA(self->derivativevalues);

    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){
        mod = &self->models[k];
        offset = mod->offset[MODEL_STATE];
        nrstates = Model_numberof(mod->modelObject)[MODEL_STATE];
        scale_inv = 1 / SCALE;
        // statevalues
        memcpy(PYDATA(Model_stateValues(mod->modelObject)),
                PYDATA(self->statevalues) + offset, 
                    nrstates * sizeof(double));
        // derivativevalues
        to = PYDATA(Model_derivativeValues(mod->modelObject));
        for(l = 0; l < nrstates; l++)
            to[l] = scale_inv * from[l + offset];
    }

    Py_RETURN_NONE;
}

static PyObject *
Simulation_ResetParameters(SimulationObject *self, PyObject *args)
{
    SimulationModel *mod;
    int k;
    const int *numberof;

    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){
        mod = &self->models[k];
        numberof = Model_numberof(mod->modelObject);
        // parametervalues
        memcpy(PYDATA(self->parametervalues) + mod->offset[MODEL_PARAMETER], 
                PYDATA(Model_parameters(mod->modelObject)),
                    numberof[MODEL_PARAMETER] * sizeof(double));
    }

    Py_RETURN_NONE;
}

static PyObject *
Simulation_UpdateModelParameters(SimulationObject *self, PyObject *args)
{
    SimulationModel *mod;
    int k;
    const int *numberof;

    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){
        mod = &self->models[k];
        numberof = Model_numberof(mod->modelObject);
        // parametervalues
        memcpy(PYDATA(Model_parameters(mod->modelObject)),
                PYDATA(self->parametervalues) + mod->offset[MODEL_PARAMETER], 
                    numberof[MODEL_PARAMETER] * sizeof(double));
    }

    Py_RETURN_NONE;
}

static PyObject *
Simulation_ValidateSimulation(SimulationObject *self, PyObject *args, PyObject *kwds){
    static char *kwlist[] = {"print", NULL};
    PyObject *print = NULL;
    int r;

    if(!PyArg_ParseTupleAndKeywords(args, kwds, "|O", kwlist, &print))
        return NULL;

    r = Simulation_CheckSharedVariables(self);
    if(PyErr_Occurred()){
        if(print && PyObject_IsTrue(print))
            PyErr_Print();
        else
            PyErr_Clear();
    }
    if(r)
        Py_RETURN_TRUE;
    else
        Py_RETURN_FALSE;
}

static PyObject *
Simulation_Simulate(SimulationObject *self, PyObject *args, PyObject *kwds){
    static char *kwlist[] = {"timevector", "timeunit", "statevalues", "derivativevalues",
                                    "parametervalues","resetstatesderivatives", "options", NULL};
    PyObject *timevector = NULL, *timeunit = NULL, *statevalues = NULL, *derivativevalues = NULL,
                *parametervalues = NULL, *resetstatesderivatives = NULL, *options = NULL;

    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|OOOOOOO", kwlist, &timevector, &timeunit, &statevalues, &derivativevalues, &parametervalues, &resetstatesderivatives, &options))
        return NULL;

    // update timevector
    if (timevector) {
        if(Simulation_SetTimeVector(self, timevector, NULL) < 0)
            return NULL;
    }
    // update timeunit
    if (timeunit) {
        if(Simulation_SetTimeUnit(self, timeunit, NULL) < 0)
            return NULL;
    }
    // update statevalues and derivatives
    if(resetstatesderivatives && PyObject_IsTrue(resetstatesderivatives)){
        if(!Simulation_ResetStatesDerivatives(self, NULL)){
            return NULL;
        }
    } else{
        if(statevalues)
            if(Simulation_SetStateValues(self, statevalues, NULL) < 0)
                return NULL;
        if(derivativevalues && derivativevalues != Py_None)
            if(Simulation_SetDerivativeValues(self, derivativevalues, NULL) < 0)
                return NULL;
    }
    // update parametervalues
    if (parametervalues) {
        if(Simulation_SetParameterValues(self, parametervalues, NULL) < 0)
            return NULL;
    }
    // update options
    if(options){
        if(Simulation_SetOptions(self, options, NULL) < 0)
            return NULL;
    }
    // init simulation
    if(Simulation_InitSimulation(self) < 0)
        return NULL;

    if(self->sundials->integrate(self->sundials, 
                                    PYSIZE(self->timevector), 
                                        PYDATA(self->timevector),
                                            PYDATA(self->statevalues),
                                                PYDATA(self->derivativevalues)) < 0)
        return NULL;
    
    // Update event data
    Simulation_updateSimulationData(self,&self->sundials->simdata);
   
    Py_RETURN_NONE;
}

static PyObject *
Simulation_ExecutionOrder(SimulationObject *self, PyObject *args)
{
    PyObject *exOrder, *name;
    int k;

    if(!self->sharedvariablescheck){
        if(!Simulation_CheckSharedVariables(self)){
            return NULL;
        }
    }
    exOrder = StringList_New(self->numberof[SIMULATION_MODEL],1);
    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){
        name = Model_name(self->models[k].modelObject);
        Py_INCREF(name);
        PyList_SetItem(exOrder, k, name);
    }
    
    return exOrder;
}

static PyObject *
Simulation_HasAlgebraicEq(SimulationObject *self, PyObject *args){
    if(self->has_algebraic_eq)
        Py_RETURN_TRUE;
    else
        Py_RETURN_FALSE;
}

static PyObject *
Simulation_DifferentialStates(SimulationObject *self, PyObject *args){
    return PyArray_Cast((PyArrayObject * ) self->idvector, NPY_BOOL);
}

static PyObject *
Simulation_SetOption(SimulationObject *self, PyObject *args, PyObject *kwds){
    static char *kwlist[] = {"option", "value", NULL};
    PyObject *option = NULL, *value = NULL;

    if(!PyArg_ParseTupleAndKeywords(args, kwds, "OO", kwlist, &option, &value))
        return NULL;

    if(!PyUnicode_Check(option)){
        PyErr_SetString(PyExc_TypeError, "'option' needs to be a string");
        return NULL;
    }

    if(PyDict_SetItem(self->options, option, value) < 0)
        return NULL;
    if(self->sundials->setoptions(self->sundials, self->options) < 0)
        return NULL;

    Py_RETURN_NONE;
}

static PyObject *
Simulation_Reduce(SimulationObject *self)
{
    int k;
    PyObject *ret, *args, *state, *timeunit, *models, *activities, *tmp;

    state = Py_BuildValue("OOO", self->eventtimedata, self->eventstatusdata, self->featuredata);

    models = PyList_New(self->numberof[SIMULATION_MODEL]);
    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++){
        tmp = self->models[k].modelObject;
        Py_INCREF(tmp);
        PyList_SetItem(models, self->models[k].originalIndex, tmp);
    }
    activities = PyList_New(self->numberof[SIMULATION_ACTIVITY]);
    for(k = 0; k < self->numberof[SIMULATION_ACTIVITY]; k++){
        tmp = self->activities[k].activityObject;
        Py_INCREF(tmp);
        PyList_SetItem(activities, k, tmp);
    }
    timeunit = Simulation_GetTimeUnit(self,NULL);
    args = Py_BuildValue("OOOOOOOO", models, activities, self->timevector, timeunit, 
                                    self->statevalues, self->derivativevalues, self->parametervalues, self->options);
    Py_DECREF(models);
    Py_DECREF(activities);
    Py_DECREF(timeunit);

    ret = Py_BuildValue("OOO", Py_TYPE(self), args, state);
    Py_DECREF(args);
    Py_DECREF(state);
    return ret;
}

static PyObject *
Simulation_SetState(SimulationObject *self, PyObject *statetuple)
{
    PyObject *tmp;

    PyObject *eventtimedata, *eventstatusdata, *featuredata;
    if(!PyArg_Parse(statetuple, "((OOO),)", &eventtimedata, &eventstatusdata, &featuredata))
        return NULL;

    if (eventtimedata != Py_None){
        if(!PyArray_CheckExact(eventtimedata) || PyArray_NDIM(eventtimedata) != 1){
            PyErr_SetString(PyExc_TypeError, "Event time data shoud be a numpy array with dimension 1");
            return NULL;
        }
    }

    if (eventstatusdata != Py_None){
        if(!PyArray_CheckExact(eventstatusdata) || PyArray_NDIM(eventstatusdata) != 2){
            PyErr_SetString(PyExc_TypeError, "Event time data shoud be a numpy array with dimension 2");
            return NULL;
        }
    }

    if (featuredata != Py_None){
        if(!PyArray_CheckExact(featuredata) || PyArray_NDIM(featuredata) != 2){
            PyErr_SetString(PyExc_TypeError, "Feature data shoud be a numpy array with dimension 2");
            return NULL;
        }
        if(PyArray_DIMS(featuredata)[1] != 0 && PyArray_DIMS(featuredata)[1] != self->numberof[SIMULATION_FEATURE]){
            PyErr_SetString(PyExc_TypeError, "Incorrect dimensionality of feature data");
            return NULL;
        }
    }

    if (eventtimedata != Py_None && eventstatusdata != Py_None){
        if(PyArray_DIMS(eventstatusdata)[1] != 0 && PyArray_DIMS(eventstatusdata)[1] != self->numberof[SIMULATION_EVENT]){
            PyErr_SetString(PyExc_TypeError, "Incorrect dimensionality of event status data");
            return NULL;
        }

        if(PyArray_DIMS(eventtimedata)[0] != PyArray_DIMS(eventstatusdata)[0]){
            PyErr_SetString(PyExc_TypeError, "Dimensons of event time and event status doesn't match");
            return NULL;
        }
    }

    tmp = self->eventtimedata;
    Py_INCREF(eventtimedata);
    self->eventtimedata = eventtimedata;
    Py_DECREF(tmp);

    tmp = self->eventstatusdata;
    Py_INCREF(eventstatusdata);
    self->eventstatusdata = eventstatusdata;
    Py_DECREF(tmp);

    tmp = self->featuredata;
    Py_INCREF(featuredata);
    self->featuredata = featuredata;
    Py_DECREF(tmp);

    Py_RETURN_NONE;
}

static PyMethodDef Simulation_methods[] = {
    {"ExecutionOrder", (PyCFunction) Simulation_ExecutionOrder, METH_NOARGS,
     "Write out the execution order"},
    {"Simulate", (PyCFunction) Simulation_Simulate, METH_VARARGS | METH_KEYWORDS,
     "Simulate simulation object"},
     {"Validate", (PyCFunction) Simulation_ValidateSimulation, METH_VARARGS | METH_KEYWORDS,
     "Validate simulation object"},
     {"ResetStatesDerivatives", (PyCFunction) Simulation_ResetStatesDerivatives, METH_NOARGS,
     "Reset simulation state and derivative values"},
     {"UpdateModelStatesDerivatives", (PyCFunction) Simulation_UpdateModelStatevalues, METH_NOARGS,
     "Update all models state and derivative values"},
     {"ResetParameters", (PyCFunction) Simulation_ResetParameters, METH_NOARGS,
     "Reset simulation parameter values"},
     {"UpdateModelParameters", (PyCFunction) Simulation_UpdateModelParameters, METH_NOARGS,
     "Update all models parameter values"},
     {"HasAlgebraicEq", (PyCFunction) Simulation_HasAlgebraicEq, METH_NOARGS,
     "Check if simulation contains algebraic equations"},
     {"DifferentialStates", (PyCFunction) Simulation_DifferentialStates, METH_NOARGS,
     "Return an array of boolean, holds the value True if respective state is a differential state"},
     {"SetOption", (PyCFunction) Simulation_SetOption, METH_VARARGS | METH_KEYWORDS,
     "Sets option for the sundials integrator"},
     {"__reduce__", (PyCFunction) Simulation_Reduce, METH_NOARGS,
     "__reduce__ function"},
     {"__setstate__", (PyCFunction) Simulation_SetState, METH_VARARGS,
     "__setstate__ function"},
    {NULL}  /* Sentinel */
};

/*
========================================================================================================================
SimulationType definition
========================================================================================================================
*/

static PyObject *
Simulation_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    SimulationObject *self;
    npy_intp dims[1];

    self = (SimulationObject *) type->tp_alloc(type, 0);
    if(self){
        // model activities
        self->models = NULL;
        self->activities = NULL;

        // timevector/featuredata/options/evettimedata/eventstatusdata
        dims[0] = 0;
        //dims[1] = 0;
        // timevector/timeunit - aka. scale
        self->timevector = PyArray_SimpleNew(1, dims, NPY_DOUBLE);
        if (!self->timevector) {
            Py_DECREF(self);
            return NULL;
        }
        self->scale = 0;
        // options
        self->options = PyDict_New();
        if (!self->options) {
            Py_DECREF(self);
            return NULL;
        }
        /*
        self->featuredata = PyArray_SimpleNew(2, dims, NPY_DOUBLE);
        if (!self->featuredata) {
            Py_DECREF(self);
            return NULL;
        }
        
        // eventtimedata
        self->eventtimedata = PyArray_SimpleNew(1, dims, NPY_DOUBLE);
        if (!self->eventtimedata) {
            Py_DECREF(self);
            return NULL;
        }
        // eventstatusdata
        self->eventstatusdata = PyArray_SimpleNew(2, dims, NPY_BOOL);
        if (!self->eventstatusdata) {
            Py_DECREF(self);
            return NULL;
        }
        */

        // Changed to inititate variables to Py_None - more clean
        // featuredata
        Py_INCREF(Py_None);
        self->featuredata = Py_None;
        // eventtimedata
        Py_INCREF(Py_None);
        self->eventtimedata = Py_None;
        // eventstatusdata
        Py_INCREF(Py_None);
        self->eventstatusdata = Py_None;
        

        // sundials memory
        self->sundials = NULL;
    }
    return (PyObject *) self;
}

static void
Simulation_initActivity(SimulationObject *self, SimulationActivity *act){
    act->offset[ACTIVITY_OUTPUT] = self->numberof[SIMULATION_OUTPUT];
    act->offset[ACTIVITY_FEATURE] = self->numberof[SIMULATION_FEATURE];

    self->numberof[SIMULATION_FEATURE] += Activity_nrFeatures(act->activityObject);
    self->numberof[SIMULATION_OUTPUT] += Activity_nrOutputs(act->activityObject);
    Activity_setNonEditable(act->activityObject);
}

static void
Simulation_initModel(SimulationObject *self, SimulationModel *mod){
    const int *mod_numberof = Model_numberof(mod->modelObject);

    mod->offset[MODEL_STATE] = self->numberof[SIMULATION_STATE];
    mod->offset[MODEL_FEATURE] = self->numberof[SIMULATION_FEATURE];
    mod->offset[MODEL_OUTPUT] = self->numberof[SIMULATION_OUTPUT];
    mod->offset[MODEL_INPUT] = self->numberof[SIMULATION_INPUT];
    mod->offset[MODEL_EVENT] = self->numberof[SIMULATION_EVENT];
    mod->offset[MODEL_PARAMETER] = self->numberof[SIMULATION_PARAMETER];
    mod->exOrder = -1;

    self->numberof[SIMULATION_STATE] += mod_numberof[MODEL_STATE];
    self->numberof[SIMULATION_FEATURE] += mod_numberof[MODEL_FEATURE];
    self->numberof[SIMULATION_OUTPUT] += mod_numberof[MODEL_OUTPUT];
    self->numberof[SIMULATION_INPUT] += mod_numberof[MODEL_INPUT];
    self->numberof[SIMULATION_EVENT] += mod_numberof[MODEL_EVENT];
    self->numberof[SIMULATION_PARAMETER] += mod_numberof[MODEL_PARAMETER];

    mod->function = Model_modelFunction(mod->modelObject);
    mod->hasoutput = mod_numberof[MODEL_OUTPUT] > 0 ? 1 : 0;
}

static int
Simulation_init(SimulationObject *self, PyObject *args, PyObject *kwds)
{
    int k, size;
    double scale;
    static char *kwlist[] = {"models", "activities", "timevector", "timeunit", "statevalues", "derivativevalues",\
                                "parametervalues", "options", NULL};
    PyObject *models = NULL, *activities = NULL, *timevector = NULL, *timeunit = NULL,\
                            *statevalues = NULL, *derivativevalues = NULL, *parametervalues = NULL, *options = NULL, *tmp;
    npy_intp dims[1];
    
    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O|OOOOOOO", kwlist, &models, &activities, &timevector, &timeunit,\
                                        &statevalues, &derivativevalues, &parametervalues, &options))
        return -1;

    self->sharedvariablescheck = 0;
    self->numberof[SIMULATION_STATE] = 0;
    self->numberof[SIMULATION_FEATURE] = 0;
    self->numberof[SIMULATION_OUTPUT] = 0;
    self->numberof[SIMULATION_INPUT] = 0;
    self->numberof[SIMULATION_PARAMETER] = 0;

    // MODEL
    size = Simulation_CheckModels(models);
    if(size <= 0){// Don't allow for empty list
        PyErr_SetString(PyExc_TypeError, "Incorrect models argument given");
        return -1;
    }

    self->models = PyMem_Calloc(size, sizeof(SimulationModel));
    self->numberof[SIMULATION_MODEL] = size;
    
    if (Model_isModel(models)){
        Py_INCREF(models);
        self->models[0].modelObject = models;
        self->models[0].originalIndex = 0;
        Simulation_initModel(self, &self->models[0]);
    }
    else{
        for(k = 0; k < size; k++){
            tmp = PyList_GetItem(models, k);
            Py_INCREF(tmp);
            self->models[k].modelObject = tmp;
            self->models[k].originalIndex = k;
            Simulation_initModel(self, &self->models[k]);
        }
    }
    // activities
    if(activities){
        size = Simulation_CheckActivities(activities);
        if(size < 0){ // Allow for empty list
            PyErr_SetString(PyExc_TypeError, "Incorrect activities argument given");
            return -1;
        }

        self->activities = PyMem_Calloc(size, sizeof(SimulationActivity));
        self->numberof[SIMULATION_ACTIVITY] = size;

        if (Activity_isActivity(activities)){
            Py_INCREF(activities);
            self->activities[0].activityObject = activities;
            Simulation_initActivity(self, &self->activities[0]);
        }
        else{
            for(k = 0; k < size; k++){
                tmp = PyList_GetItem(activities, k);
                Py_INCREF(tmp);
                self->activities[k].activityObject = tmp;
                Simulation_initActivity(self, &self->activities[k]);
            }
        }
    } else {
        self->numberof[SIMULATION_ACTIVITY] = 0;
    }

    // model simulation data
    // statevalues
    dims[0] = self->numberof[SIMULATION_STATE];
    self->statevalues = PyArray_ZEROS(1, dims, NPY_DOUBLE, 0);
    if (!self->statevalues)
        return -1;
    // derivativevalues
    self->derivativevalues = PyArray_ZEROS(1, dims, NPY_DOUBLE, 0);
    if (!self->derivativevalues)
        return -1;
    // idvector
    self->idvector = PyArray_ZEROS(1, dims, NPY_DOUBLE, 0);
    if (!self->idvector)
        return -1;
    dims[0] = self->numberof[SIMULATION_PARAMETER];
    // parametervalues
    self->parametervalues = PyArray_ZEROS(1, dims, NPY_DOUBLE, 0);
    if (!self->parametervalues)
        return -1;

    // names lists
    // statenames
    self->statenames = StringList_New(self->numberof[SIMULATION_STATE],1);
    if (!self->statenames)
        return -1;
    // featurenames
    self->featurenames = StringList_New(self->numberof[SIMULATION_FEATURE],1);
    if (!self->featurenames)
        return -1;
    // featureunits
    self->featureunits = StringList_New(self->numberof[SIMULATION_FEATURE],1);
    if (!self->featureunits)
        return -1;
    // outputnames
    self->outputnames = StringList_New(self->numberof[SIMULATION_OUTPUT],1);
    if (!self->outputnames)
        return -1;
    // inputnames
    self->inputnames = StringList_New(self->numberof[SIMULATION_INPUT],1);
    if (!self->inputnames)
        return -1;
    // parameternames
    self->parameternames = StringList_New(self->numberof[SIMULATION_PARAMETER],1);
    if (!self->parameternames)
        return -1;
    // eventnames
    self->eventnames = StringList_New(self->numberof[SIMULATION_EVENT],1);
    if (!self->eventnames)
        return -1;

    // Update simulation attribute names
    if(Simulation_SetAttributeNames(self) < 0)
        return -1;

    // Constructor values
    // timevector
    if (timevector) {
        if(Simulation_SetTimeVector(self, timevector, NULL) < 0)
            return -1;
    }
    // timeunit
    if (timeunit) {
        scale = Simulation_getTimeScale(self,timeunit);
        if(scale < 0)
            return -1;
        else
            self->scale = scale;
    }
    else
        self->scale = 1; // default is seconds
    if(Simulation_updateModelActivityScale(self) < 0)
        return -1;
    // parametervalues
    if(parametervalues){
        if(Simulation_SetParameterValues(self, parametervalues, NULL) < 0)
            return -1;
    } else{
        tmp = Simulation_ResetParameters(self, NULL); // return Py_None
        Py_DECREF(tmp);
    }
    // Reset states derivative, need to be done first since both 
    // states/derivatives are effected
    tmp = Simulation_ResetStatesDerivatives(self,NULL); // return Py_None
    Py_DECREF(tmp);
    // statevalues
    if(statevalues){
        if(Simulation_SetStateValues(self, statevalues, NULL) < 0)
            return -1;
    }
    // derivativevalues
    if(derivativevalues){
        if(Simulation_SetDerivativeValues(self, derivativevalues, NULL) < 0)
            return -1;
    }

    // check algebraic states and set id vector
    self->has_algebraic_eq = 0;
    Simulation_idVectorAlgebraicEqs(self);

    // input output buffer/execution list
    self->outputbuffer = PyMem_Calloc(self->numberof[SIMULATION_OUTPUT], sizeof(double));
    self->inputptr = PyMem_Calloc(self->numberof[SIMULATION_INPUT], sizeof(double *));
    self->inputmap = PyMem_Calloc(self->numberof[SIMULATION_INPUT], sizeof(int));

    // sundials object
    if(self->has_algebraic_eq){
        self->sundials = IDA55_init((void *) self, 
                                        self->numberof[SIMULATION_STATE],
                                            self->numberof[SIMULATION_EVENT],
                                                self->numberof[SIMULATION_FEATURE],
                                                    PYDATA(self->idvector),
                                                        self->options);
        if(!self->sundials)
            return -1;
    } else{
        self->sundials = CVODE55_init((void *) self, 
                                        self->numberof[SIMULATION_STATE], 
                                            self->numberof[SIMULATION_EVENT], 
                                                self->numberof[SIMULATION_FEATURE],
                                                    self->options);
        if(!self->sundials)
            return -1;
    }
    // sundials options
    if(options){
        if(Simulation_SetOptions(self, options, NULL) < 0)
            return -1;
    }
    
    return 0;
}

static void
Simulation_dealloc(SimulationObject *self)
{
    int k;
    // models
    for(k = 0; k < self->numberof[SIMULATION_MODEL]; k++)
        Py_XDECREF(self->models[k].modelObject);
    PyMem_Free(self->models);
    // activities
    for(k = 0; k < self->numberof[SIMULATION_ACTIVITY]; k++)
        Py_XDECREF(self->activities[k].activityObject);
    PyMem_Free(self->activities);

    Py_XDECREF(self->timevector);
    Py_XDECREF(self->options);
    Py_XDECREF(self->statevalues);
    Py_XDECREF(self->derivativevalues);
    Py_XDECREF(self->idvector);
    Py_XDECREF(self->parametervalues);
    Py_XDECREF(self->featuredata);
    Py_XDECREF(self->eventtimedata);
    Py_XDECREF(self->eventstatusdata);
    Py_XDECREF(self->featurenames);
    Py_XDECREF(self->featureunits);
    Py_XDECREF(self->outputnames);
    Py_XDECREF(self->inputnames);
    Py_XDECREF(self->parameternames);
    Py_XDECREF(self->statenames);
    Py_XDECREF(self->eventnames);
    PyMem_Free(self->outputbuffer);
    PyMem_Free(self->inputptr);
    PyMem_Free(self->inputmap);

    // sundials object
    self->sundials->free(self->sundials);

    Py_TYPE(self)->tp_free((PyObject *) self);
}

static PyTypeObject SimulationType = {
    PyVarObject_HEAD_INIT(NULL, 0)
    .tp_name = "MeDigiT._Simulation.Simulation",
    .tp_doc = "Simulation Object",
    .tp_basicsize = sizeof(SimulationObject),
    .tp_itemsize = 0,
    .tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,
    .tp_new = Simulation_new,
    .tp_init = (initproc) Simulation_init,
    .tp_dealloc = (destructor) Simulation_dealloc,
    .tp_getset = Simulation_getsetters,
    .tp_methods = Simulation_methods,
};

/*
==========================================================================================
C_API FUNCTIONS
==========================================================================================
*/

/*
==========================================================================================
Simulation module definition
==========================================================================================
*/

static PyModuleDef Simulation = {
    PyModuleDef_HEAD_INIT,
    .m_name = "MeDigiT._Simulation",
    .m_doc = "Python interface to Simuation.",
    .m_size = -1
};

PyMODINIT_FUNC
PyInit__Simulation(void)
{
    PyObject *m, *timeScales, *tmp;
    TimeScale *iterator;

    m = PyModule_Create(&Simulation);
    if (m == NULL)
        return NULL;

    // Simulation type
    if (PyType_Ready(&SimulationType) < 0){
        Py_DECREF(m);
        return NULL;
    }

    Py_INCREF(&SimulationType);

   if (PyModule_AddObject(m, "Simulation", (PyObject *) &SimulationType) < 0){
        Py_DECREF(&SimulationType);
        Py_DECREF(m);
        return NULL;
    }

    import_array();
    import_StringList();
    import_Activity();
    import_Models();

    // Timescale list
    iterator = timeScaleData;
    timeScales = StringList_New(0,1);
    if(!timeScales){
        Py_DECREF(&SimulationType);
        Py_DECREF(m);
        return NULL;
    }
    while(iterator->name){
        tmp = PyUnicode_FromString(iterator->name);
        if(PyList_Append(timeScales,tmp) < 0){
            Py_DECREF(timeScales);
            Py_DECREF(tmp);
            Py_DECREF(&SimulationType);
            Py_DECREF(m);
            return NULL;
        }
        Py_DECREF(tmp);
        iterator++;
    }

    if (PyModule_AddObject(m, "TimeScales", timeScales) < 0){
        Py_DECREF(timeScales);
        Py_DECREF(&SimulationType);
        Py_DECREF(m);
        return NULL;
    }

    return m;
}
