#ifndef _SIMULATION_H
#define _SIMULATION_H

#ifdef _SIMULATION_C
/*
==========================================================================================
Simulation defines
==========================================================================================
*/
// MODEL MACROS
#define STATEVECTOR &statevalues[offset[MODEL_STATE]]
#define DERIVATEVECTOR &derivativevalues[offset[MODEL_STATE]]
#define RESIDUALVECTOR &RESvector[offset[MODEL_STATE]]
#define PARAMETERVECTOR &parametervalues[offset[MODEL_PARAMETER]]
#define FEATUREVECTOR &featurevector[offset[MODEL_FEATURE]]
#define OUTPUTVECTOR &self->outputbuffer[offset[MODEL_OUTPUT]]
#define INPUTVECTOR &self->inputptr[offset[MODEL_INPUT]]
#define EVENTVECTOR &eventvector[offset[MODEL_EVENT]]
#define EVENTSTATUS &eventstatus[offset[MODEL_EVENT]]
#define MODELFUNCTION (mod->function)
#define HASOUTPUT (mod->hasoutput)
#define SCALE (mod->scale)

// ACTIVITY MACROS
#define A_OUTPUTVECTOR &self->outputbuffer[offset[ACTIVITY_OUTPUT]]
#define A_FEATUREVECTOR &featurevector[offset[ACTIVITY_FEATURE]]
#define A_SCALE (act->scale)

#define ATTRIBUTE_FEATURE_NAME 0
#define ATTRIBUTE_FEATURE_UNIT 1
#define ATTRIBUTE_OUTPUT_NAME 2
#define ATTRIBUTE_INPUT_NAME 3
#define ATTRIBUTE_PARAMETER_NAME 4
#define ATTRIBUTE_STATE_NAME 5
#define ATTRIBUTE_EVENT_NAME 6
#define ATTRIBUTE_STATE_VALUE 7
#define ATTRIBUTE_DERIVATE_VALUE 8
#define ATTRIBUTE_PARAMETER_VALUE 9


/*
==========================================================================================
Structure definitions
==========================================================================================
*/

typedef struct{
    PyObject *modelObject;
    ModelFunction *function;
    int offset[6];
    int hasoutput;
    int exOrder;
    double scale;
    int originalIndex;
}SimulationModel;

typedef struct{
    PyObject *activityObject;
    int offset[2];
    double scale;
}SimulationActivity;

typedef struct{
    PyObject_HEAD
    // models and activities
    SimulationModel *models;
    SimulationActivity *activities;
    // model simulation data
    PyObject *statevalues;
    PyObject *derivativevalues;
    PyObject *idvector;
    PyObject *parametervalues;
    // model attributes
    PyObject *featurenames;
    PyObject *featureunits;
    PyObject *outputnames;
    PyObject *inputnames;
    PyObject *parameternames;
    PyObject *statenames;
    PyObject *eventnames;
    // simulation data
    SUNDIALSObject *sundials;
    PyObject *timevector;
    PyObject *options;
    PyObject *featuredata;
    PyObject *eventtimedata;
    PyObject *eventstatusdata; 
    int sharedvariablescheck;
    int numberof[8];
    double *outputbuffer;
    double **inputptr;
    int *inputmap;
    double scale;
    int has_algebraic_eq;
}SimulationObject;

/*
==========================================================================================
Function declaration
==========================================================================================
*/
extern int IDA55(void *simObject, int nrtimes, double *timevector, int nrstates, double *statevalues, double *derivativevalues, double *idvector, int nrevents, double **eventtimes, int **eventstatus, int *nreventshappened, PyObject *optionsPY);
static void Simulation_initModel(SimulationObject *self, SimulationModel *mod);
static void Simulation_initActivity(SimulationObject *self, SimulationActivity *act);

#else
/*
==========================================================================================
C_API
==========================================================================================
*/
#endif
#endif
