#!python3
# Script to install MeDigiT package

try:
  from setuptools_scm import integration
  integration.find_files = lambda p: []
except ImportError:
  pass

from setuptools import setup, Extension, find_packages
from setuptools.command.develop import develop
from setuptools.command.build_py import build_py
import numpy
import os
import glob
from distutils.ccompiler import new_compiler
from shutil import copy,rmtree,move
from sysconfig import get_paths
import sys

packages = find_packages(".")
package_dir={"": "."}

MeDigiTFolder = "MeDigiT"
include = [numpy.get_include(),
            os.path.join(MeDigiTFolder,"sundials-5_5_0","src"),
              os.path.join(MeDigiTFolder,"include"),
                os.path.join(MeDigiTFolder,"_Activity"),
                  os.path.join(MeDigiTFolder,"_Models"),
                    os.path.join(MeDigiTFolder,"_Simulation"),
                      os.path.join(MeDigiTFolder,"_StringList")]

_Activity = Extension("MeDigiT._Activity",
              include_dirs = include,
              sources = [os.path.join(MeDigiTFolder,"_Activity","_Activity.c")])
_Models = Extension("MeDigiT._Models",
              include_dirs = include + [os.path.join(MeDigiTFolder,"_Models","Models","include")],
              sources = [os.path.join(MeDigiTFolder,"_Models","_Models.c")])

sundialsfiles = glob.glob(os.path.join(MeDigiTFolder,"sundials-5_5_0", "**/*.c"), recursive = True)
_Simulation = Extension("MeDigiT._Simulation",
              include_dirs = include + [os.path.join(MeDigiTFolder,"sundials-5_5_0","src","include")],
              sources = [os.path.join(MeDigiTFolder,"_Simulation","_Simulation.c")] + sundialsfiles)
_StringList = Extension("MeDigiT._StringList",
              include_dirs = include,
              sources = [os.path.join(MeDigiTFolder,"_StringList","_StringList.c")])

##################################################################################################
# Compile ExtraFormulas Library
##################################################################################################


##################################################################################################
# Compile ExtraFormulas Library and copy necessary files
##################################################################################################
def _preProccessing():
  # sundials
  curdir = os.getcwd()
  os.chdir(os.path.join(MeDigiTFolder,"sundials-5_5_0"))
  copy("src/SUNDIALSflags55.h","../Models/include/SUNDIALSflags55.h")

  # ExtraFormulas
  os.chdir("../ExtraFormulas")
  cfiles = glob.glob("*.c")
  # compile library
  compiler = new_compiler()
  objects = compiler.compile(sources=cfiles,output_dir="Objects",include_dirs=[get_paths()["include"]])
  compiler.create_static_lib(objects=objects,output_libname='ExtraFormulas',output_dir="../Models/include")
  copy("mathaddon.h","../Models/include/mathaddon.h")
  copy("kineticformulas.h","../Models/include/kineticformulas.h")
  # remove object files
  rmtree('Objects')

  # copy Models.h
  os.chdir("../")
  copy("_Models/_Models.h","Models/include/_Models.h")

  os.chdir(curdir)

class custom_develop(develop):
  """Custom build command."""
  def run(self):
    _preProccessing()
    super().run()

class custom_build(build_py):
  """Custom build command."""
  def run(self):
    _preProccessing()
    super().run()


setup(name="MeDigiT",
      version="1.0.0",
      description="Python interface for the MeDigiT toolbox",
      author="Gustav Magnusson",
      author_email="bogustavmagnusson@gmail.com",
      url="liu.se",
      #install_requires = ["numpy"], # numpy needs to be installed before MeDigiT since its path is used above
      cmdclass={'build_py' : custom_build, 'develop' : custom_develop},
      packages=packages,
      package_dir=package_dir,
      include_package_data=True,
      package_data={'MeDigiT.Models': ['include/*']},
      ext_modules=[_Activity,_Models,_Simulation,_StringList])

