import MeDigiT as MDT
import numpy as np

print("We need to import models to be used")
Glucose = MDT.importModel("Glucose")
Insulin = MDT.importModel("Insulin")

print("We can than create models")
M1 = Glucose()
M2 = Insulin()
A1 = MDT.Activity("s")

print("Current statevalues in module objects:")
print("M1: " + str(M1.statenames))
print("M1: " + str(M1.statevalues))
print("M2: " + str(M2.statenames))
print("M2: " + str(M2.statevalues) + "\n")

print("Current derivativevalues in module objects:")
print("M1: " + str(M1.derivativevalues))
print("M2: " + str(M2.derivativevalues) + "\n")

print("Parameters in the module objects:")
print("M1: " + str(M1.parameternames))
print("M1: " + str(M1.parametervalues))
print("M2: " + str(M2.parameternames))
print("M2: " + str(M2.parametervalues) + "\n")

print("Features the module objects can simulate:")
print("M1: " + str(M1.featurenames))
print("M2: " + str(M2.featurenames))
print("\nObserve attribute name change from features -> features\n")

print("Units of the features:")
print("M1: " + str(M1.featureunits))
print("M2: " + str(M2.featureunits) + "\n")

print("Default time unit used in model: (ns = nanosecond, μs = microsecond, ms = millisecond, s = second, m = minute, h = hour, d = day, y = year")
print("M1: " + str(M1.timeunit))
print("M2: " + str(M2.timeunit) + "\n")

print("Outputs the module objects compute when simulating:")
print("M1: " + str(M1.outputnames))
print("M2: " + str(M2.outputnames) + "\n")

print("Input the module objects need when simulating:")
print("M1: " + str(M1.inputnames))
print("M2: " + str(M2.inputnames) + "\n")

print("Current outputs and features in Activity object")
print("A1 outputs: " + str(A1.outputnames))
print("A1 feature names: " + str(A1.featurenames))
print("A1 feature units: " + str(A1.featureunits))
print()

print("We can add outputs and features to Activities, currently they can be of four different types:\n" + 
		"MeDigiT.CONSTANT, MeDigiT.PIECEWISE_CONSTANT, MeDigiT.PIECEWISE_LINEAR, MeDigiT.CUBIC_INTERPOLATE, MeDigiT.CUBIC_INTERPOLATE_SLOPE\n")


A1.AddOutput(type = MDT.CUBIC_SPLINE, name = "ACTIVITY_CUBIC", tvalues = [0,1,2], fvalues = [0,1,0], unit = "T", feature = True)
A1.AddOutput(type = MDT.CONSTANT, name = "Activity_CONSTANT", fvalues = 100, feature = False) # if unit is left out it is assumed to be dimensionless, i.e unit 1

print("A1 outputs: " + str(A1.outputnames))
print("A1 feature names: " + str(A1.featurenames))
print("\nObserve attribute name change from features -> features\n")
print("A1 feature units: " + str(A1.featureunits))

print("\nWe can create a simulation with these two module objects:")
A1.AddOutput(type = MDT.PIECEWISE_CONSTANT, name = "ACTIVITY_IN", unit= "kg", tvalues = [0,1,2], fvalues = [0,0,1,2])

sim = MDT.Simulation(models = [M1,M2], activities = [A1])
print("A simulation can be validate by either calling 'Validate' or 'ExecutionOrder\n")

print("Validate: " + str(sim.Validate(print=True))) # return true or false whether or not simulation is valid, can print out error
sim.ExecutionOrder() # return an error set if simulation not valid or prints out the exectution order of models if simulation is valid
print()

print("Using the simulation object we can simulate the module objects:\n")
time = [i*0.5 for i in range(10)]
options = {"calcic" : True, "maxnumsteps" : 10000}
sim.Simulate(timevector = time, timeunit = 's', resetstatesderivatives=True, options = options)

print("Feature data and units:")
print(sim.featurenames)
print(sim.featureunits)
print()
print(sim.featuredata)
print()

print("State and derivative values:")
print(sim.statenames)
print(sim.statevalues)
print()
print(sim.derivativevalues)
print()

print("Note that state values are not updated automatically:")
print("M1: " + str(M1.statevalues))
print("M2: " + str(M2.statevalues) + "\n")

print("But need to update them explicitley by running 'UpdateModelStatevalues':")
sim.UpdateModelStatesDerivatives()
print("M1: " + str(M1.statevalues))
print("M2: " + str(M2.statevalues) + "\n")

print("Current derivativevalues in simulation object:")
print(sim.derivativevalues)
print()
print("Current derivativevalues in module objects:")
print("M1: " + str(M1.derivativevalues))
print("M2: " + str(M2.derivativevalues) + "\n")

print("The events are save:")
print("Event names: " + str(sim.eventnames))
print("Event times: " + str(sim.eventtimes))
print("Event status: " + str(sim.eventstatus) + "\n")

print("Simulation has algebraic equations: " + str(sim.HasAlgebraicEq()))
print(sim.DifferentialStates())
print(M1.DifferentialStates())
print(M2.DifferentialStates())
print(M1.HasAlgebraicEq())
print(M2.HasAlgebraicEq())

